#include <iostream>
using namespace std;

#include "game.h"

// program entry point
int main(int argc, char** argv) {

	Game* g;

	g = new Game(argc, argv);
	g->init();
	g->run();
	return g->return_code;

} // main


// initialization mananger
void Game::init() {
	init_rules();
	init_io();
	init_gfx();
	init_snd();
	init_physics();
}

// game specific initialization
void Game::init_rules () {
	// TODO: add some game specific setup logic here
}

// setup SDL systems
void Game::init_io ()	{

	// misc setup
	SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO);
    atexit(SDL_Quit); // register SDL Parachute for crashes
	SDL_WM_SetCaption("game", NULL); // give our window a name (default icon)

	// NOTE: SDL_Init can be called with SDL_RESIZABLE to allow the user to
	// resize the window.  However, this would require resetting projections
	// and such whenever it happened. KISS.

	// set opengl attributes (is this necesary on all systems?)
    SDL_GL_SetAttribute(SDL_GL_RED_SIZE,        5);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,      5);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,       5);
#ifdef __APPLE__
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      32);
#else
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      16);
#endif
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,    1);

	// create framebuffer
	const SDL_VideoInfo* vidinfo = SDL_GetVideoInfo();
	SDL_Surface* screen = NULL;
	Uint32 fs = show_fullscreen?SDL_FULLSCREEN:0;
	screen = SDL_SetVideoMode(win_width,win_height,vidinfo->vfmt->BitsPerPixel,SDL_OPENGL|fs);
	if(!screen) {
		cerr << "Couldn't set video mode!" << endl;
		cerr << SDL_GetError() << endl;
		exit(1);
	}

} // init_io

// setup GL and load visual assets
void Game::init_gfx() {

	// viewport
    glViewport(0,0,win_width,win_height);

	// matricies
	glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45, (float)win_width/win_height, 1, 100);
	glMatrixMode(GL_TEXTURE);
		// unflip textures from SDL_image
		glRotatef(180,0,0,1);
		glScalef(-1,1,1);
	glMatrixMode(GL_MODELVIEW);
	
	// clear color
	glClearColor(1,1,1,0);

	// misc GL state
	glEnable(GL_NORMALIZE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_COLOR_MATERIAL);
	glShadeModel(GL_SMOOTH);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // most common setting

	// texturing init
	glPixelStorei(GL_UNPACK_ALIGNMENT,4);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	SDL_Surface* s = NULL; // reused to load each texture

	try {
		// "unit_circle.png" (no alpha channel)
		s = IMG_Load("unit_circle.png");
		if(!s) { throw "unit_circle.png";}
		glGenTextures(1, &tex_something);
		glBindTexture(GL_TEXTURE_2D,tex_something);
		gluBuild2DMipmaps(GL_TEXTURE_2D,4,s->w, s->h,
			GL_RGB,GL_UNSIGNED_BYTE,s->pixels); // RGB
		SDL_FreeSurface(s);

		// "soft_mask.png" (with alpha channel)
		s = IMG_Load("soft_mask.png");
		if(!s) { throw "soft_mask.png";}
		glGenTextures(1, &tex_mask);
		glBindTexture(GL_TEXTURE_2D,tex_mask);
		gluBuild2DMipmaps(GL_TEXTURE_2D,4,s->w, s->h,
			GL_RGBA,GL_UNSIGNED_BYTE,s->pixels); // RGBA
		SDL_FreeSurface(s);

	} catch (const char* filename) {
		cerr << "Failed to loading texture " << filename << endl;
		cerr << SDL_GetError() << endl;
		exit(1);
	}	

	// this would be a good place to load any fancy geometry asssets required 

} // init_gfx

// setup audio device and load audio assets
void Game::init_snd() {

	// setup audio mixer
    Mix_OpenAudio(44100, AUDIO_S16, 2, 4096); // common settings
    // only if needed: Mix_AllocateChannels(128);

	try {

		// load and start playing music
		music_ambient = Mix_LoadMUS("ambient.ogg");
		if(!music_ambient) { throw "ambient.ogg"; }
		Mix_PlayMusic(music_ambient,-1); // loop forever

		// load chunks (samples)
		chunk_rattle = Mix_LoadWAV("rattle.ogg");
		if(!music_ambient) { throw "rattle.ogg"; }

	}  catch (const char* filename) {
		cerr << "Failed to loading audio " << filename << endl;
		cerr << SDL_GetError() << endl;
		exit(1);
	}	

} // init_snd()

void Game::init_physics() {

	// grab time
	ticks = SDL_GetTicks()-1; // -1 ensures SOME time passes for first frame's motion

	// create containers
	world = dWorldCreate();
	space = dHashSpaceCreate(0);
	contact_jgroup = dJointGroupCreate(0);

	// gravity
	dWorldSetGravity(world,0,0,-9.8);

	// game bodies
	dMass m; // same mass properties for all objects
	dMassSetBox(&m,1,1,1,1);
	for(int i=0; i<some; i++) {
		dBodyID b = dBodyCreate(world);
        dGeomID g = dCreateBox(space,1,1,1);
		dBodySetMass(b,&m);
        dGeomSetBody(g,b);
        dBodySetPosition(b,0,0,1+1.1*i);
		dBodySetData(b,NULL); // set the user pointer to something helps us identify the body
		bodies.push_back(b);

		if(i>0) {
			dJointID j = dJointCreateBall(world,0);
			dJointAttach(j,bodies[i],bodies[i-1]);
			dJointSetBallAnchor(j,0,0,1.1*i+0.5);
		}
    } // for some;

	// infinite ground plane (has no body!)
	dCreatePlane(space,0,0,1,0);

} // init_physics

// game update manager
void Game::run() {

	SDL_Event e;

	// main loop -- one screen draw for each iteration
	while(running) {

		// handle all available events before continuing (could lag?)		
		while(SDL_PollEvent(&e)) {
			switch(e.type) {
				case SDL_KEYDOWN:
				case SDL_KEYUP:
					handle_KeyboardEvent(e.key);
					break;
				case SDL_MOUSEMOTION:
					handle_MouseMotionEvent(e.motion);
					break;
				case SDL_MOUSEBUTTONDOWN:
				case SDL_MOUSEBUTTONUP:
					handle_MouseButtonEvent(e.button);
					break;
				case SDL_USEREVENT:
					handle_UserEvent(e.user);
					break;
				case SDL_QUIT: // generated by WM
					quit();
				// no default, ignore all other events quickly
			} // switch
		} // while PollEvent


		calc_dt();

		tick_physics();
		tick_rules();

		render();

	} // while running

} // run

// set config flags from arguments
void Game::process_args() {
	// TODO: do something with these
} // process_args

// examine the wall clock time
void Game::calc_dt() {
	Uint32 now = SDL_GetTicks();
	dt = (now-ticks)/1000.0;
	ticks = now;
}

// draw pretty stuff on the screen
void Game::render() {

	// reset stuff
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();

	// example smooth camera trick
    const dReal* target = dBodyGetPosition(bodies[some-1]);
	cout << dt << endl; // XXX
    static float ex,ey,ez;
    static float lx,ly,lz;
    static float dece = 0.005;
    ex -= dece*(ex-target[0]);
    ey -= dece*(ey-target[1]);
    ez -= dece*(ez-target[2]);
    static float decl = 0.05;
    lx -= decl*(lx-target[0]);
    ly -= decl*(ly-target[1]);
    lz -= decl*(lz-target[2]);
    gluLookAt(0*ex+0.2,ey-15,ez+5, lx,ly,lz, 0,0,1);


	// place the light (up from here)
	const float up[] = {0,0,1,0};
	glLightfv(GL_LIGHT0,GL_POSITION,up);

	// decorate gound plane
    glPushAttrib(GL_ALL_ATTRIB_BITS);  // save GL state
		glColor3f(1,1,1);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_DEPTH_TEST);
		glBindTexture(GL_TEXTURE_2D, tex_something);
		glBegin(GL_QUADS);
			glNormal3f(0,0,1);
			glTexCoord2i(0,0);
			glVertex3f(-5,-5,0);
			glTexCoord2i(1,0);
			glVertex3f(+5,-5,0);
			glTexCoord2i(1,1);
			glVertex3f(+5,+5,0);
			glTexCoord2i(0,1);
			glVertex3f(-5,+5,0);
		glEnd();
    glPopAttrib(); // recall old state (old color and texturing in this case)
    
	// draw each physics body as a visual cube (with "reflections") 
	glPushAttrib(GL_ALL_ATTRIB_BITS);
		for (unsigned i=0; i<bodies.size(); i++) {
			const float* mat = bodyTransform(bodies[i]);
			glColor3f(1,0,0);
			glPushMatrix();
				glMultMatrixf(mat);
				glutSolidCube(1);
			glPopMatrix();
			//glDisable(GL_LIGHTING);
			glColor3f(0,0,0);
			glPushMatrix();
				glScalef(1,1,0);
				glMultMatrixf(mat);
				glutSolidCube(1); 
			glPopMatrix();
			
		} // for bodies
    glPopAttrib();

    // draw HUD (screen space 2D graphics)
	glPushAttrib(GL_ALL_ATTRIB_BITS);
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glPushMatrix(); // about to clobber nice perspective tranform!
			glDisable(GL_LIGHTING);
			glDisable(GL_DEPTH_TEST); // dont let world objects stick thru the screen
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, tex_mask); // has interesting alpha
			glLoadIdentity();
			glEnable(GL_BLEND);
			glColor3f(1,1,1); // with GL_COLOR_MATERIAL on we can tint the texture
			glBegin(GL_QUADS);
				glNormal3f(0,0,1);
				glTexCoord2i(0,0);
				glVertex3f(-1,-1,0);
				glTexCoord2i(1,0);
				glVertex3f(+1,-1,0);
				glTexCoord2i(1,1);
				glVertex3f(+1,+1,0);
				glTexCoord2i(0,1);
				glVertex3f(-1,+1,0);
			glEnd();
			glDisable(GL_TEXTURE_2D);
			glColor4f(1,1,1,0.2);
			glTranslatef(-0.8,-0.8,0); 
			glScalef(1,ar(),1);
			glRotatef(180*T(), 0,0,1); // spin with time
			glutSolidCube(0.2);
		glPopMatrix();
	glPopAttrib();

	// show our picture to the user (oh, won't they be proud)
    SDL_GL_SwapBuffers();

} // render

// tell ODE to update all physical bodies
void Game::tick_physics() {

	dJointGroupEmpty(contact_jgroup);
	dSpaceCollide(space,0,&Game::cb_near);
    dWorldStep(world,dt);

} // tick_physics

// high level game logic updater
void Game::tick_rules() {

	// TODO: apply game rules here
	if(dAreConnected(bodies[some-1],0)) {
		SDL_Event r;
		r.type = SDL_USEREVENT;
		r.user.code = 0;
		SDL_PushEvent(&r);
	}

} // tick_rules

// builds GL matrix from given bodies position/rotation
// returns pointer to local buffer, modified by subsequent calls
const GLfloat*  Game::bodyTransform(dBodyID b) {

	static GLfloat mat[16]; // to be returned

	const dReal* pos = dBodyGetPosition(b);
    const dReal* rot = dBodyGetRotation(b);
   
	// 0 applies no perspective transform
	// 1 preserves homogeneous coordinate
    mat[0] = rot[0];
    mat[1] = rot[4];
    mat[2] = rot[8];
    mat[3] = 0;
    mat[4] = rot[1];
    mat[5] = rot[5];
    mat[6] = rot[9];
    mat[7] = 0;
    mat[8] = rot[2];
    mat[9] = rot[6];
    mat[10] = rot[10];
    mat[11] = 0;
    mat[12] = pos[0];
    mat[13] = pos[1];
    mat[14] = pos[2];
    mat[15] = 1;

	return mat;
} // applyBodyTransform

// interpret key up and down events
void Game::handle_KeyboardEvent(SDL_KeyboardEvent e) {

	// TODO: handle them!
	// e.keysym.sym is like SDLK_{keyname}
	// e.state is either SDL_PRESSED or SDL_RELEASED
	if(e.keysym.sym == SDLK_ESCAPE && e.state == SDL_PRESSED) {
		quit(); // tell the main loop to finish up after next frame
	}

	if(e.keysym.sym == SDLK_r && e.state == SDL_PRESSED) {
		SDL_Event r;
		r.type = SDL_USEREVENT;
		r.user.code = 0;
		SDL_PushEvent(&r);
	}

} // handle_KeyBoardEvent


// interpret mouse movement events
void Game::handle_MouseMotionEvent(SDL_MouseMotionEvent e) {

	// TODO: handle them!
	// e.x and e.y are Uint16 with window coords

} // handle_MouseMotionEvent

// interpret mouse button up and down events
void Game::handle_MouseButtonEvent(SDL_MouseButtonEvent e) {

	// TODO: handle them!
	// e.button is SDL_BUTTON_LEFT, SDL_BUTTON_MIDDLE, SDL_BUTTON_RIGHT
	// e.state is either SDL_PRESSED or SDL_RELEASED

} // handle_MoouseButtonEvent


// handle events generated by other game logic
// SDL has a nice event system... let's hijack it
void Game::handle_UserEvent(SDL_UserEvent e) {

	// TODO: add some good game logic here
	cout << "handle_UserEvent:" << endl;
	cout << "\t e.code = " << e.code << endl;
	switch(e.code) {
		case 0:
			Mix_PlayChannel(-1,chunk_rattle,0);
			for(unsigned int i=0; i<bodies.size(); i++) {
				dBodySetPosition(bodies[i],0,0,1+1.1*i);
				dBodySetLinearVel(bodies[i],0,0,0);
				dBodySetAngularVel(bodies[i],0,0,0);
			}
			break;
	}

} // handle_UserEvent

// handle collisions between physical bodies (complex function!)
void Game::cb_near(void *data, dGeomID o1, dGeomID o2) {
    if (dGeomIsSpace (o1) || dGeomIsSpace (o2)) {

      // colliding a space with something
      dSpaceCollide2 (o1,o2,data,&Game::cb_near);
      // collide all geoms internal to the space(s)
      if (dGeomIsSpace (o1)) dSpaceCollide ((dSpaceID)o1,data,&Game::cb_near);
      if (dGeomIsSpace (o2)) dSpaceCollide ((dSpaceID)o2,data,&Game::cb_near);

    } else {

		// colliding two non-space geoms, so generate contact
		// points between o1 and o2
        dBodyID b1 = dGeomGetBody(o1);
		dBodyID b2 = dGeomGetBody(o2);

		dContact contacts[g->max_contacts];
		int n = dCollide(o1,o2, g->max_contacts, &contacts[0].geom, sizeof(dContact));
		for(int i=0; i<n; i++) {
			contacts[i].surface.mode = 0; // | more
			contacts[i].surface.mu = 10;
			dJointID c = dJointCreateContact(g->world,g->contact_jgroup,&contacts[i]);
			dJointAttach(c,b1,b2);
		} // for n
    }

} // cb_near

// THE END