#ifndef _GAME_H
#define _GAME_H

#include <vector>
using namespace std;

#include <SDL.h>
#include <SDL_mixer.h>
#include <SDL_image.h>
#include <ode/ode.h>

#ifdef __APPLE__
#	include <GLUT/glut.h>
#else
#	include <glut.h>
#endif

// "singleton" instance
class Game;
Game* g;

// a high-level container class for game state and logic
class Game {

// in a serious game much of this would be marked private
public:

	/////////////
	// METHODS //
	/////////////

	Game(int argc, char** argv)
		: argc(argc),argv(argv),win_width(800),win_height(600),show_fullscreen(false),dt(0),running(true),return_code(0)
		{
			g = this;
			this->process_args();
		};

	void process_args(); // examine command line arguments

	void init(); // calls init_*

	void init_rules(); // high level game setup
	void init_io(); // SDL setup
	void init_gfx(); // GL setup, visual asset preparation (using SDL_image)
	void init_snd(); // sound setup, audio asset preparation (using SDL_mixer)
	void init_physics(); // ODE setup, basic world setup

	void run(); // main event loop (handle, tick, draw)

	// handle common SDL events
	void handle_KeyboardEvent(SDL_KeyboardEvent);
	void handle_MouseMotionEvent(SDL_MouseMotionEvent);
	void handle_MouseButtonEvent(SDL_MouseButtonEvent);
	void handle_UserEvent(SDL_UserEvent); // we borrow SDL's event system

	void calc_dt(); // update Ticks and dt using SDL_GetTicks

	void tick_physics(); // ODE world step
	void tick_rules(); // per frame logical rules update
	void render(); // fill screen with fun stuff

	static void cb_near(void*,dGeomID,dGeomID);

	// utility functions

	void quit(int code=0) {running=false; return_code=code;}
	double ar() {return (double)win_width/win_height;}
	static const GLfloat* bodyTransform(dBodyID);

	static double r(double max=1.0,double min=0.0) {
		return ((max-min)*(double)rand()/(double)RAND_MAX)+min;
	}

	////////////////
	// PROPERTIES //
	////////////////

	// MISC
	int argc;
	char **argv;
	Uint32 ticks; // last result of SDL_GetTicks
	double dt; // time in seconds since last update
	inline double FPS() {return dt;} // frames per second
	inline double SPF() {return 1.0/dt;} // seconds per frame
	inline double T() {return (double)ticks/1000.0;} // seconds per frame
	bool running; // flag for event loop
	int return_code; // value to return to OS from main

	// RULES
	// TODO: add some game specific stuff here!
	static const int some = 12;

	// IO
	int win_width;
	int win_height;
	bool show_fullscreen;

	// GRAPHICS
	GLuint tex_something;
	GLuint tex_mask;

	// SOUND
	Mix_Music* music_ambient;
	Mix_Chunk* chunk_rattle;

	// PHYSICS
	dWorldID world;
	dSpaceID space;
	vector<dBodyID> bodies;
	dJointGroupID contact_jgroup;
	static const int max_contacts = 8;


}; // class Game

#endif